/*	tik.c

	Temporal Image Kentucky/Kontainer

	By Henry Dietz, June 2016
*/
		
#include "tik.h"


char	*myname;	/* For error messages, name of this program */

FILE	*TDCIfile = 0;

double	e_begin = 0;
char	*e_noisefile = 0;
double	e_fps = 0;		/* Frames Per Second */
double	e_gamma = 0;		/* Encoding gamma (0 means not set) */
int	e_n = 0;
char	*e_outfile = 0;
char	*e_outtik = "pnm.tik";
char	*e_outimage = "tik%05d.jpg";
double	e_percent = 4.55;
double	e_quality = 75;
double	e_t = 0;		/* Shutter time, Tv in seconds */
int	e_what = WHAT_UNSPEC;
int	e_interactive = 0;	/* Interactive mode? */
time_t	e_update;
int	e_maketype = MAKE_UNSPEC;


double
atofrac(char *s)
{
	/* Atof, but allowing fractions with 1/ */
	if ((s[0] == '1') && (s[1] == '/')) {
		return(1.0 / atof(&(s[2])));
	}
	return(atof(s));
}

int
main(register int argc,
register char **argv)
{
	register char *s;
	register int i, n;
	register char *fname = 0;
	register double now;
	time_t real_time = (e_update = time(0));

	/* Set interactive mode based on stderr */
	e_interactive = isatty(2);

	/* Process command line... */
	myname = argv[0];
	if (argc < 2) {
usage:
		fprintf(stderr,
			"Usage: %s {exposure_settings}\n"
			"-a#       Set Tv by shutter angle; Tv=(angle/360)/FPS\n"
			"-b#       Begin exposure time in seconds (default 0); needs .tik input\n"
			"-eNAME    Error model is NAME; blank creates from video of a constant scene\n"
			"-f#       Frames per second, FPS (default 24)\n"
			"-g#       Gamma of encoded data (1.0 linear default; 2.2 typical JPEG)\n"
			"-i#       Interactive progess update seconds-1 (default %d, 0 means never, -i toggles)\n"
			"-n#       Number of frames to encode/decode (default 1 output, all input)\n"
			"-oNAME    Output filename is NAME (default %s)\n"
			"-p#       Minimum %% probability same to factor (default %0.2f; std. dev. are 32, 5, 0.3)\n"
			"-q#       Quality %% (default %0.0f; for JPEG output or TIK encoding)\n"
			"-t#       Shutter speed in seconds, Tv (default to 1/FPS or 1/60s)\n"
			"-v        Input is a video to be converted to .tik TDCI output\n"
			"FILENAME  Process file as specified by earlier options, .tik assumed to be TDCI\n",
			myname,
			e_interactive,
			e_outtik,
			e_percent,
			e_quality);
		exit(ERROR_USAGE);
	}

	for (i=1; i<argc; ++i) {
		if (*argv[i] == '-') {
			switch (*(argv[i]+1)) {
			case 'a':
				e_t = atofrac(argv[i]+2);
				if (e_fps == 0) {
					ERROR(ERROR_ARG,
					      "cannot set shutter angle before -f to set FPS\n");
				}
				if (e_t <= 0) {
					ERROR(ERROR_ARG,
					      "shutter angle must be greater than 0, not %s\n",
					      (argv[i]+2));
				}
				e_t = (e_t / (e_fps * 360.0));
				break;
			case 'b':
				e_begin = atofrac(argv[i]+2);
				e_what |= WHAT_TIK;
				break;
			case 'e':
				e_noisefile = (argv[i]+2);
				if (*e_noisefile) {
					/* Read error model file */
					uint8 *p;

					if ((0 > (PNMfd = open(e_noisefile, O_RDONLY))) ||
					    (0 == (p = PNMreader())) ||
					    (PNMx != 256) ||
					    (PNMy != 256) ||
					    (PNMmaxval > 255)) {
						ERROR(ERROR_READ,
						      "could not read error model file %s\n",
						      e_noisefile);
					}
					if (wavetype != waveNOISE) {
						WARN("file %s used, but not marked as TIK NOISE\n",
						     e_noisefile);
					}

					memcpy(&(PNMerr[0][0][0]), p, (256*256*3));
					close(PNMfd);
					free(p);

					/* Only makes sense for encoding */
					e_what |= WHAT_VIDEO;
				} else {
					/* Force create error model file */
					e_what = WHAT_ERRMOD;
				}
				break;
			case 'f':
				e_fps = atofrac(argv[i]+2);
				break;
			case 'g':
				e_gamma = atofrac(argv[i]+2);
				break;
			case 'i':
				if (*(argv[i]+2) == 0) {
					e_interactive = !e_interactive;
				} else {
					e_interactive = atoi(argv[i]+2);
					if (e_interactive < 1) e_interactive = 1;
				}
				break;
			case 'n':
				if (*(argv[i]+2) == 0) {
					e_n = 0;
				} else {
					e_n = atoi(argv[i]+2);
					if (e_n < 1) e_n = 1;
				}
				break;
			case 'o':
				e_outfile = (argv[i]+2);

				/* Guess type based on end of filename... */
				if (e_maketype == MAKE_UNSPEC) {
					register char *p = argv[i] + 2;
					register int len = strlen(p);

					switch (len) {
					case 1:
						if (*p != '-') break;
						/* Fall through... */
					case 0:
						/* PPM stream to stdout for "-o-" or "-o" */
						e_outfile = "-";
						e_maketype = MAKE_PPM;
						if (e_what != WHAT_ERRMOD) e_what |= WHAT_TIK;
						break;
					case 2:
					case 3:
						/* Too short to say... */
						break;
					default:
						/* Does it end in something we know? */
						p += (len - 4);
						if (p[0] == '.') {
							if (((p[1] == 'J') || (p[1] == 'j')) &&
							    ((p[2] == 'P') || (p[2] == 'p')) &&
							    ((p[3] == 'G') || (p[3] == 'g'))) {
								e_maketype = MAKE_JPEG;
								e_what |= WHAT_TIK;
							}
							if (((p[1] == 'P') || (p[1] == 'p')) &&
							    ((p[2] == 'P') || (p[2] == 'p') || (p[2] == 'N') || (p[2] == 'n')) &&
							    ((p[3] == 'M') || (p[3] == 'm'))) {
								e_maketype = MAKE_PPM;
								if (e_what != WHAT_ERRMOD) e_what |= WHAT_TIK;
							}
						}
					}
				}
				break;
			case 'p':
				e_percent = atofrac(argv[i]+2);
				if ((e_percent < 0) || (e_percent > 100)) {
					ERROR(ERROR_ARG,
					      "%s is not valid; must be between -p0 and -p100\n",
					      argv[i]);
				}
				e_what |= WHAT_VIDEO;
				break;
			case 'q':
				e_quality = atofrac(argv[i]+2);
				if ((e_quality < 0) || (e_quality > 100)) {
					ERROR(ERROR_ARG,
					      "%s is not valid; must be between -q0 and -q100\n",
					      argv[i]);
				}
				break;
			case 't':
				e_t = atofrac(argv[i]+2);
				break;
			case 'v':
				e_what |= WHAT_VIDEO;
				break;
			default:
				goto usage;
			}
		} else {
			/* What are we supposed to do with this file?
			   Can't be nothing or more than one thing.
			*/
			register char *p = argv[i];
			register int dottik = 0;

			/* Guess based on end of filename... */
			while (*p) ++p;
			p -= 4;
			if ((p[0] == '.') &&
			    ((p[1] == 'T') || (p[1] == 't')) &&
			    ((p[2] == 'I') || (p[2] == 'i')) &&
			    ((p[3] == 'K') || (p[3] == 'k'))) {
				/* Force encoding exposures */
				e_what = WHAT_TIK;
				dottik = 1;
			} else {
				if (e_what == WHAT_UNSPEC) e_what = WHAT_VIDEO;
			}

//			if (e_t && (e_fps == 0)) e_fps = 1.0 / e_t;
//			if (e_fps && (e_t == 0)) e_t = 1.0 / e_fps;
			switch (e_what) {
			case WHAT_ERRMOD:
				if (e_n == 0) e_n = E_N_MAX;
				if (e_outfile == 0) e_outfile = e_outtik;
				INFO("creating error model \"%s\" from \"%s\"\n",
				     e_outfile,
				     argv[i]);
				start_ffmpeg(argv[i]);
				PNMnoise(e_outfile);
				stop_ffmpeg();
				break;
			case WHAT_VIDEO:
				if (e_n == 0) e_n = E_N_MAX;
				if (e_outfile == 0) e_outfile = e_outtik;
				if (e_noisefile == 0) {
					/* No error model specified; make one */
					register int x, y;
					for (y=0; y<256; ++y) {
						for (x=0; x<256; ++x) {
							int z = 255 - ((x > y) ? (x-y) : (y-x));
							if (z < 0) z = 0;
							z -= ((255 - z) * (255 - z));
							if (z < 0) z = 0;
							PNMerr[x][y][0] = z;
							PNMerr[x][y][1] = z;
							PNMerr[x][y][2] = z;
						}
					}
				}
				start_ffmpeg(argv[i]);
				PNMtdci(e_outfile);
				stop_ffmpeg();
				break;
			case WHAT_TIK:
				if (e_outfile == 0) e_outfile = e_outimage;
				now = ((e_begin < 0) ? 0 : e_begin);
				if (e_n <= 1) {
					/* One frame only */
					e_n = 1;
					if (e_t == 0) e_t = ((e_fps > 0) ? (1.0/e_fps) : (1.0/60));
					if (e_fps == 0) e_fps = e_t;
				} else {
					/* Video sequence */
					if (e_fps == 0) e_fps = 24;
					if (e_t == 0) e_t = 1.0 / e_fps;
				}

				/* Create exposures */
				TIKrenderFile(dottik, argv[i]);
				for (n=0; n<e_n; ++n) {
					TIKrender(now, now+e_t);
					now += (1.0 / e_fps);
				}

				/* Simultaneously render all frames */
				TIKrenderSimul();

				/* Done with this file */
				break;
			default:
				ERROR(ERROR_ARG,
				      "what to do with %s? %s%s%s\n",
				      argv[i],
				      ((e_what & WHAT_ERRMOD) ? " Make error model?" : ""),
				      ((e_what & WHAT_VIDEO) ? " Encode as TIK?" : ""),
				      ((e_what & WHAT_TIK) ? " Make exposures?" : ""));
			}
		}
	}

	INFO("completed in %d seconds\n", ((int)(time(0) - real_time)));
	return(0);
}
